










PyDoc_STRVAR(termios__doc__,
"This module provides an interface to the Posix calls for tty I/O control.\n\
For a complete description of these calls, see the Posix or Unix manual\n\
pages. It is only available for those Unix versions that support Posix\n\
termios style tty I/O control.\n\
\n\
All functions in this module take a file descriptor fd as their first\n\
argument. This can be an integer file descriptor, such as returned by\n\
sys.stdin.fileno(), or a file object, such as sys.stdin itself.");

static PyObject *TermiosError;

static int fdconv(PyObject* obj, void* p)
{
	int fd;

	fd = PyObject_AsFileDescriptor(obj);
	if (fd >= 0) {
		*(int*)p = fd;
		return 1;
	}
	return 0;
}

PyDoc_STRVAR(termios_tcgetattr__doc__,
"tcgetattr(fd) -> list_of_attrs\n\
\n\
Get the tty attributes for file descriptor fd, as follows:\n\
[iflag, oflag, cflag, lflag, ispeed, ospeed, cc] where cc is a list\n\
of the tty special characters (each a string of length 1, except the items\n\
with indices VMIN and VTIME, which are integers when these fields are\n\
defined).  The interpretation of the flags and the speeds as well as the\n\
indexing in the cc array must be done using the symbolic constants defined\n\
in this module.");

static PyObject *
termios_tcgetattr(PyObject *self, PyObject *args)
{
	int fd;
	struct termios mode;
	PyObject *cc;
	speed_t ispeed, ospeed;
	PyObject *v;
	int i;
	char ch;

	if (!PyArg_ParseTuple(args, "O&:tcgetattr", 
			      fdconv, (void*)&fd))
		return NULL;

	if (tcgetattr(fd, &mode) == -1)
		return PyErr_SetFromErrno(TermiosError);

	ispeed = cfgetispeed(&mode);
	ospeed = cfgetospeed(&mode);

	cc = PyList_New(NCCS);
	if (cc == NULL)
		return NULL;
	for (i = 0; i < NCCS; i++) {
		ch = (char)mode.c_cc[i];
		v = PyString_FromStringAndSize(&ch, 1);
		if (v == NULL)
			goto err;
		PyList_SetItem(cc, i, v);
	}

	
	if ((mode.c_lflag & ICANON) == 0) {
		v = PyInt_FromLong((long)mode.c_cc[VMIN]);
		if (v == NULL)
			goto err;
		PyList_SetItem(cc, VMIN, v);
		v = PyInt_FromLong((long)mode.c_cc[VTIME]);
		if (v == NULL)
			goto err;
		PyList_SetItem(cc, VTIME, v);
	}

	if (!(v = PyList_New(7)))
		goto err;

	PyList_SetItem(v, 0, PyInt_FromLong((long)mode.c_iflag));
	PyList_SetItem(v, 1, PyInt_FromLong((long)mode.c_oflag));
	PyList_SetItem(v, 2, PyInt_FromLong((long)mode.c_cflag));
	PyList_SetItem(v, 3, PyInt_FromLong((long)mode.c_lflag));
	PyList_SetItem(v, 4, PyInt_FromLong((long)ispeed));
	PyList_SetItem(v, 5, PyInt_FromLong((long)ospeed));
	PyList_SetItem(v, 6, cc);
	if (PyErr_Occurred()){
		Py_DECREF(v);
		goto err;
	}
	return v;
  err:
	Py_DECREF(cc);
	return NULL;
}

PyDoc_STRVAR(termios_tcsetattr__doc__,
"tcsetattr(fd, when, attributes) -> None\n\
\n\
Set the tty attributes for file descriptor fd.\n\
The attributes to be set are taken from the attributes argument, which\n\
is a list like the one returned by tcgetattr(). The when argument\n\
determines when the attributes are changed: termios.TCSANOW to\n\
change immediately, termios.TCSADRAIN to change after transmitting all\n\
queued output, or termios.TCSAFLUSH to change after transmitting all\n\
queued output and discarding all queued input. ");

static PyObject *
termios_tcsetattr(PyObject *self, PyObject *args)
{
	int fd, when;
	struct termios mode;
	speed_t ispeed, ospeed;
	PyObject *term, *cc, *v;
	int i;

	if (!PyArg_ParseTuple(args, "O&iO:tcsetattr", 
			      fdconv, &fd, &when, &term))
		return NULL;
	if (!PyList_Check(term) || PyList_Size(term) != 7) {
		PyErr_SetString(PyExc_TypeError, 
			     "tcsetattr, arg 3: must be 7 element list");
		return NULL;
	}

	
	if (tcgetattr(fd, &mode) == -1)
		return PyErr_SetFromErrno(TermiosError);
	mode.c_iflag = (tcflag_t) PyInt_AsLong(PyList_GetItem(term, 0));
	mode.c_oflag = (tcflag_t) PyInt_AsLong(PyList_GetItem(term, 1));
	mode.c_cflag = (tcflag_t) PyInt_AsLong(PyList_GetItem(term, 2));
	mode.c_lflag = (tcflag_t) PyInt_AsLong(PyList_GetItem(term, 3));
	ispeed = (speed_t) PyInt_AsLong(PyList_GetItem(term, 4));
	ospeed = (speed_t) PyInt_AsLong(PyList_GetItem(term, 5));
	cc = PyList_GetItem(term, 6);
	if (PyErr_Occurred())
		return NULL;

	if (!PyList_Check(cc) || PyList_Size(cc) != NCCS) {
		PyErr_Format(PyExc_TypeError, 
			"tcsetattr: attributes[6] must be %d element list",
			     NCCS);
		return NULL;
	}

	for (i = 0; i < NCCS; i++) {
		v = PyList_GetItem(cc, i);

		if (PyString_Check(v) && PyString_Size(v) == 1)
			mode.c_cc[i] = (cc_t) * PyString_AsString(v);
		else if (PyInt_Check(v))
			mode.c_cc[i] = (cc_t) PyInt_AsLong(v);
		else {
			PyErr_SetString(PyExc_TypeError, 
     "tcsetattr: elements of attributes must be characters or integers");
			return NULL;
		}
	}

	if (cfsetispeed(&mode, (speed_t) ispeed) == -1)
		return PyErr_SetFromErrno(TermiosError);
	if (cfsetospeed(&mode, (speed_t) ospeed) == -1)
		return PyErr_SetFromErrno(TermiosError);
	if (tcsetattr(fd, when, &mode) == -1)
		return PyErr_SetFromErrno(TermiosError);

	Py_INCREF(Py_None);
	return Py_None;
}

PyDoc_STRVAR(termios_tcsendbreak__doc__,
"tcsendbreak(fd, duration) -> None\n\
\n\
Send a break on file descriptor fd.\n\
A zero duration sends a break for 0.25-0.5 seconds; a nonzero duration\n\
has a system dependent meaning.");

static PyObject *
termios_tcsendbreak(PyObject *self, PyObject *args)
{
	int fd, duration;

	if (!PyArg_ParseTuple(args, "O&i:tcsendbreak", 
			      fdconv, &fd, &duration))
		return NULL;
	if (tcsendbreak(fd, duration) == -1)
		return PyErr_SetFromErrno(TermiosError);

	Py_INCREF(Py_None);
	return Py_None;
}

PyDoc_STRVAR(termios_tcdrain__doc__,
"tcdrain(fd) -> None\n\
\n\
Wait until all output written to file descriptor fd has been transmitted.");

static PyObject *
termios_tcdrain(PyObject *self, PyObject *args)
{
	int fd;

	if (!PyArg_ParseTuple(args, "O&:tcdrain", 
			      fdconv, &fd))
		return NULL;
	if (tcdrain(fd) == -1)
		return PyErr_SetFromErrno(TermiosError);

	Py_INCREF(Py_None);
	return Py_None;
}

PyDoc_STRVAR(termios_tcflush__doc__,
"tcflush(fd, queue) -> None\n\
\n\
Discard queued data on file descriptor fd.\n\
The queue selector specifies which queue: termios.TCIFLUSH for the input\n\
queue, termios.TCOFLUSH for the output queue, or termios.TCIOFLUSH for\n\
both queues. ");

static PyObject *
termios_tcflush(PyObject *self, PyObject *args)
{
	int fd, queue;

	if (!PyArg_ParseTuple(args, "O&i:tcflush", 
			      fdconv, &fd, &queue))
		return NULL;
	if (tcflush(fd, queue) == -1)
		return PyErr_SetFromErrno(TermiosError);

	Py_INCREF(Py_None);
	return Py_None;
}

PyDoc_STRVAR(termios_tcflow__doc__,
"tcflow(fd, action) -> None\n\
\n\
Suspend or resume input or output on file descriptor fd.\n\
The action argument can be termios.TCOOFF to suspend output,\n\
termios.TCOON to restart output, termios.TCIOFF to suspend input,\n\
or termios.TCION to restart input.");

static PyObject *
termios_tcflow(PyObject *self, PyObject *args)
{
	int fd, action;

	if (!PyArg_ParseTuple(args, "O&i:tcflow", 
			      fdconv, &fd, &action))
		return NULL;
	if (tcflow(fd, action) == -1)
		return PyErr_SetFromErrno(TermiosError);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyMethodDef termios_methods[] =
{
	{"tcgetattr", termios_tcgetattr, 
	 METH_VARARGS, termios_tcgetattr__doc__},
	{"tcsetattr", termios_tcsetattr, 
	 METH_VARARGS, termios_tcsetattr__doc__},
	{"tcsendbreak", termios_tcsendbreak, 
	 METH_VARARGS, termios_tcsendbreak__doc__},
	{"tcdrain", termios_tcdrain, 
	 METH_VARARGS, termios_tcdrain__doc__},
	{"tcflush", termios_tcflush, 
	 METH_VARARGS, termios_tcflush__doc__},
	{"tcflow", termios_tcflow, 
	 METH_VARARGS, termios_tcflow__doc__},
	{NULL, NULL}
};




static struct constant {
	char *name;
	long value;
} termios_constants[] = {
	
	{"B0", B0},
	{"B50", B50},
	{"B75", B75},
	{"B110", B110},
	{"B134", B134},
	{"B150", B150},
	{"B200", B200},
	{"B300", B300},
	{"B600", B600},
	{"B1200", B1200},
	{"B1800", B1800},
	{"B2400", B2400},
	{"B4800", B4800},
	{"B9600", B9600},
	{"B19200", B19200},
	{"B38400", B38400},
	{"B57600", B57600},
	{"B115200", B115200},
	{"B230400", B230400},
	{"CBAUDEX", CBAUDEX},

	
	{"TCSANOW", TCSANOW},
	{"TCSADRAIN", TCSADRAIN},
	{"TCSAFLUSH", TCSAFLUSH},

	
	{"TCIFLUSH", TCIFLUSH},
	{"TCOFLUSH", TCOFLUSH},
	{"TCIOFLUSH", TCIOFLUSH},

	
	{"TCOOFF", TCOOFF},
	{"TCOON", TCOON},
	{"TCIOFF", TCIOFF},
	{"TCION", TCION},

	
	{"IGNBRK", IGNBRK},
	{"BRKINT", BRKINT},
	{"IGNPAR", IGNPAR},
	{"PARMRK", PARMRK},
	{"INPCK", INPCK},
	{"ISTRIP", ISTRIP},
	{"INLCR", INLCR},
	{"IGNCR", IGNCR},
	{"ICRNL", ICRNL},
	{"IUCLC", IUCLC},
	{"IXON", IXON},
	{"IXANY", IXANY},
	{"IXOFF", IXOFF},
	{"IMAXBEL", IMAXBEL},

	
	{"OPOST", OPOST},
	{"OLCUC", OLCUC},
	{"ONLCR", ONLCR},
	{"OCRNL", OCRNL},
	{"ONOCR", ONOCR},
	{"ONLRET", ONLRET},
	{"OFILL", OFILL},
	{"OFDEL", OFDEL},
	{"NLDLY", NLDLY},
	{"CRDLY", CRDLY},
	{"TABDLY", TABDLY},
	{"BSDLY", BSDLY},
	{"VTDLY", VTDLY},
	{"FFDLY", FFDLY},

	
	{"NL0", NL0},
	{"NL1", NL1},
	{"CR0", CR0},
	{"CR1", CR1},
	{"CR2", CR2},
	{"CR3", CR3},
	{"TAB0", TAB0},
	{"TAB1", TAB1},
	{"TAB2", TAB2},
	{"TAB3", TAB3},
	{"XTABS", XTABS},
	{"BS0", BS0},
	{"BS1", BS1},
	{"VT0", VT0},
	{"VT1", VT1},
	{"FF0", FF0},
	{"FF1", FF1},

	
	{"CSIZE", CSIZE},
	{"CSTOPB", CSTOPB},
	{"CREAD", CREAD},
	{"PARENB", PARENB},
	{"PARODD", PARODD},
	{"HUPCL", HUPCL},
	{"CLOCAL", CLOCAL},
	{"CIBAUD", CIBAUD},
	{"CRTSCTS", (long)CRTSCTS},

	
	{"CS5", CS5},
	{"CS6", CS6},
	{"CS7", CS7},
	{"CS8", CS8},

	
	{"ISIG", ISIG},
	{"ICANON", ICANON},
	{"XCASE", XCASE},
	{"ECHO", ECHO},
	{"ECHOE", ECHOE},
	{"ECHOK", ECHOK},
	{"ECHONL", ECHONL},
	{"ECHOCTL", ECHOCTL},
	{"ECHOPRT", ECHOPRT},
	{"ECHOKE", ECHOKE},
	{"FLUSHO", FLUSHO},
	{"NOFLSH", NOFLSH},
	{"TOSTOP", TOSTOP},
	{"PENDIN", PENDIN},
	{"IEXTEN", IEXTEN},

	
	{"VINTR", VINTR},
	{"VQUIT", VQUIT},
	{"VERASE", VERASE},
	{"VKILL", VKILL},
	{"VEOF", VEOF},
	{"VTIME", VTIME},
	{"VMIN", VMIN},
	
	{"VSWTC", VSWTC},
	{"VSWTCH", VSWTCH},
	{"VSTART", VSTART},
	{"VSTOP", VSTOP},
	{"VSUSP", VSUSP},
	{"VEOL", VEOL},
	{"VREPRINT", VREPRINT},
	{"VDISCARD", VDISCARD},
	{"VWERASE", VWERASE},
	{"VLNEXT", VLNEXT},
	{"VEOL2", VEOL2},


	{"B460800", B460800},
	{"CBAUD", CBAUD},
	{"CDEL", CDEL},
	{"CDSUSP", CDSUSP},
	{"CEOF", CEOF},
	{"CEOL", CEOL},
	{"CEOL2", CEOL2},
	{"CEOT", CEOT},
	{"CERASE", CERASE},
	{"CESC", CESC},
	{"CFLUSH", CFLUSH},
	{"CINTR", CINTR},
	{"CKILL", CKILL},
	{"CLNEXT", CLNEXT},
	{"CNUL", CNUL},
	{"COMMON", COMMON},
	{"CQUIT", CQUIT},
	{"CRPRNT", CRPRNT},
	{"CSTART", CSTART},
	{"CSTOP", CSTOP},
	{"CSUSP", CSUSP},
	{"CSWTCH", CSWTCH},
	{"CWERASE", CWERASE},
	{"EXTA", EXTA},
	{"EXTB", EXTB},
	{"FIOASYNC", FIOASYNC},
	{"FIOCLEX", FIOCLEX},
	{"FIONBIO", FIONBIO},
	{"FIONCLEX", FIONCLEX},
	{"FIONREAD", FIONREAD},
	{"IBSHIFT", IBSHIFT},
	{"INIT_C_CC", INIT_C_CC},
	{"IOCSIZE_MASK", IOCSIZE_MASK},
	{"IOCSIZE_SHIFT", IOCSIZE_SHIFT},
	{"NCC", NCC},
	{"NCCS", NCCS},
	{"NSWTCH", NSWTCH},
	{"N_MOUSE", N_MOUSE},
	{"N_PPP", N_PPP},
	{"N_SLIP", N_SLIP},
	{"N_STRIP", N_STRIP},
	{"N_TTY", N_TTY},
	{"TCFLSH", TCFLSH},
	{"TCGETA", TCGETA},
	{"TCGETS", TCGETS},
	{"TCSBRK", TCSBRK},
	{"TCSBRKP", TCSBRKP},
	{"TCSETA", TCSETA},
	{"TCSETAF", TCSETAF},
	{"TCSETAW", TCSETAW},
	{"TCSETS", TCSETS},
	{"TCSETSF", TCSETSF},
	{"TCSETSW", TCSETSW},
	{"TCXONC", TCXONC},
	{"TIOCCONS", TIOCCONS},
	{"TIOCEXCL", TIOCEXCL},
	{"TIOCGETD", TIOCGETD},
	{"TIOCGICOUNT", TIOCGICOUNT},
	{"TIOCGLCKTRMIOS", TIOCGLCKTRMIOS},
	{"TIOCGPGRP", TIOCGPGRP},
	{"TIOCGSERIAL", TIOCGSERIAL},
	{"TIOCGSOFTCAR", TIOCGSOFTCAR},
	{"TIOCGWINSZ", TIOCGWINSZ},
	{"TIOCINQ", TIOCINQ},
	{"TIOCLINUX", TIOCLINUX},
	{"TIOCMBIC", TIOCMBIC},
	{"TIOCMBIS", TIOCMBIS},
	{"TIOCMGET", TIOCMGET},
	{"TIOCMIWAIT", TIOCMIWAIT},
	{"TIOCMSET", TIOCMSET},
	{"TIOCM_CAR", TIOCM_CAR},
	{"TIOCM_CD", TIOCM_CD},
	{"TIOCM_CTS", TIOCM_CTS},
	{"TIOCM_DSR", TIOCM_DSR},
	{"TIOCM_DTR", TIOCM_DTR},
	{"TIOCM_LE", TIOCM_LE},
	{"TIOCM_RI", TIOCM_RI},
	{"TIOCM_RNG", TIOCM_RNG},
	{"TIOCM_RTS", TIOCM_RTS},
	{"TIOCM_SR", TIOCM_SR},
	{"TIOCM_ST", TIOCM_ST},
	{"TIOCNOTTY", TIOCNOTTY},
	{"TIOCNXCL", TIOCNXCL},
	{"TIOCOUTQ", TIOCOUTQ},
	{"TIOCPKT", TIOCPKT},
	{"TIOCPKT_DATA", TIOCPKT_DATA},
	{"TIOCPKT_DOSTOP", TIOCPKT_DOSTOP},
	{"TIOCPKT_FLUSHREAD", TIOCPKT_FLUSHREAD},
	{"TIOCPKT_FLUSHWRITE", TIOCPKT_FLUSHWRITE},
	{"TIOCPKT_NOSTOP", TIOCPKT_NOSTOP},
	{"TIOCPKT_START", TIOCPKT_START},
	{"TIOCPKT_STOP", TIOCPKT_STOP},
	{"TIOCSCTTY", TIOCSCTTY},
	{"TIOCSERCONFIG", TIOCSERCONFIG},
	{"TIOCSERGETLSR", TIOCSERGETLSR},
	{"TIOCSERGETMULTI", TIOCSERGETMULTI},
	{"TIOCSERGSTRUCT", TIOCSERGSTRUCT},
	{"TIOCSERGWILD", TIOCSERGWILD},
	{"TIOCSERSETMULTI", TIOCSERSETMULTI},
	{"TIOCSERSWILD", TIOCSERSWILD},
	{"TIOCSER_TEMT", TIOCSER_TEMT},
	{"TIOCSETD", TIOCSETD},
	{"TIOCSLCKTRMIOS", TIOCSLCKTRMIOS},
	{"TIOCSPGRP", TIOCSPGRP},
	{"TIOCSSERIAL", TIOCSSERIAL},
	{"TIOCSSOFTCAR", TIOCSSOFTCAR},
	{"TIOCSTI", TIOCSTI},
	{"TIOCSWINSZ", TIOCSWINSZ},
	{"TIOCTTYGSTRUCT", TIOCTTYGSTRUCT},

	
	{NULL, 0}
};


PyMODINIT_FUNC
PyInit_termios(void)
{
	PyObject *m;
	struct constant *constant = termios_constants;

	m = Py_InitModule4("termios", termios_methods, termios__doc__,
                           (PyObject *)NULL, PYTHON_API_VERSION);
	if (m == NULL)
		return;

	if (TermiosError == NULL) {
		TermiosError = PyErr_NewException("termios.error", NULL, NULL);
	}
	Py_INCREF(TermiosError);
	PyModule_AddObject(m, "error", TermiosError);

	while (constant->name != NULL) {
		PyModule_AddIntConstant(m, constant->name, constant->value);
		++constant;
	}
}
